from django.utils import six

try:
    import importlib
except ImportError:
    from django.utils import importlib  # noqa


def import_attribute(path):
    assert isinstance(path, six.string_types)
    pkg, attr = path.rsplit('.', 1)
    ret = getattr(importlib.import_module(pkg), attr)
    return ret



from django.contrib import admin

from .models import Activity, Notice


class NoticeAdmin(admin.ModelAdmin):
    date_hierarchy = 'created_at'
    list_display = (
        'activity',
        'user',
        'created_at',
        'read_at',
        'archived_at')
    raw_id_fields = (
        'user',
        'activity')
    list_filter = (
        'created_at',)


class ActivityAdmin(admin.ModelAdmin):
    date_hierarchy = 'created_at'
    list_display = (
        'type',
        'created_at')
    list_filter = (
        'created_at',)


admin.site.register(Notice, NoticeAdmin)
admin.site.register(Activity, ActivityAdmin)



DATABASE_ENGINE = 'sqlite3'

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': ':memory:',
    }
}

INSTALLED_APPS = [
    'django.contrib.contenttypes',
    'django.contrib.auth',
    'django.contrib.admin',

    'actistream',
    'actistream.tests'
]

USE_TZ = True

SECRET_KEY = 'psst'

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [
            # insert your TEMPLATE_DIRS here
        ],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                # Insert your TEMPLATE_CONTEXT_PROCESSORS here or use this
                # list if you haven't customized them:
                'django.contrib.auth.context_processors.auth',
                'django.template.context_processors.debug',
                'django.template.context_processors.i18n',
                'django.template.context_processors.media',
                'django.template.context_processors.static',
                'django.template.context_processors.tz',
                'django.contrib.messages.context_processors.messages',
            ],
        },
    },
]



from django.test import TestCase
from django.contrib.auth import get_user_model
from django.core import mail

from actistream.models import Notice, Activity

from . import activities
from .models import Comment, Article


User = get_user_model()


class ActivityTestCase(TestCase):

    def setUp(self):
        self.user = User.objects.create(
            email='john@doe.com',
            username='john')
        self.article = Article.objects.create()
        self.comment = Comment.objects.create(
            article=self.article,
            user=self.user)
        self.activity = activities.CommentPosted.create(
            target=self.article,
            actor=self.user,
            action_object=self.comment)

    def test_notice(self):
        notice_recipients = [self.user]
        Notice.objects.send(
            self.activity,
            notice_recipients)
        msg = mail.outbox[0]
        self.assertEquals(msg.subject, 'Comment posted')
        assert msg.to == ['john@doe.com']
        assert msg.content_subtype == 'html'

    def test_for_target(self):
        qs = Activity.objects.for_target(self.article)
        assert qs.count() == 1
        assert qs.filter(pk=self.activity.pk).exists()

    def test_for_action_object(self):
        qs = Activity.objects.for_action_object(self.comment)
        assert qs.count() == 1
        assert qs.filter(pk=self.activity.pk).exists()

    def test_for_action_objects(self):
        qs = Activity.objects.for_action_objects(
            Comment.objects.all())
        assert qs.count() == 1
        assert qs.filter(pk=self.activity.pk).exists()

    def test_fetch_related(self):
        # TODO: Actually test something
        Activity.objects.fetch_related(Activity.objects.all())



from __future__ import absolute_import

from collections import defaultdict

from django.conf import settings
from django.db import models
from django.db.models import Q, F
from django.core.exceptions import ObjectDoesNotExist
from django.utils.translation import ugettext_lazy as _
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.utils.encoding import python_2_unicode_compatible

from actistream import registry


class ActivityManager(models.Manager):

    def _filter_flags(self, qs, flags):
        if flags:
            qry = None
            for mask in flags:
                q = Q(**{'flags': F('flags').bitor(mask)})
                if not qry:
                    qry = q
                else:
                    qry = qry | q
            qs = qs.filter(qry)
        return qs

    def for_action_objects(self, action_objects):
        ct = ContentType.objects.get_for_model(action_objects.model)
        activities = Activity.objects \
            .filter(action_object_ct=ct,
                    action_object_id__in=action_objects.values_list(
                        'id',
                        flat=True))
        return activities

    def for_action_object(self, action_object):
        ct = ContentType.objects.get_for_model(action_object)
        activities = Activity.objects \
            .filter(action_object_ct=ct,
                    action_object_id=action_object.pk)
        return activities

    def for_target(self, target, types=None, flags=None):
        """
        flags: [1, 2, 12] == 1 or 2 or (8 and 4)
        """
        ct = ContentType.objects.get_for_model(target)
        activities = Activity.objects \
            .filter(target_ct=ct,
                    target_id=target.pk)
        if types:
            activities = activities.filter(type__in=[t.id for t in types])
        if flags:
            activities = self._filter_flags(activities, flags)
        return activities.order_by('-created_at')

    def fetch_related(self, qs):
        """
        Excludes activities with incomplete (deleted) relations
        """
        ct_to_obj_ids = defaultdict(set)
        fields = ['actor', 'target', 'action_object']
        # Collect objects to fetch
        for activity in qs:
            for field in fields:
                ct_id = getattr(activity, field + '_ct_id')
                obj_id = getattr(activity, field + '_id')
                if ct_id and obj_id:
                    ct_to_obj_ids[ct_id].add(obj_id)
        # Fetch objects
        ct_to_obj_id_to_obj = {}
        for ct_id, obj_ids in ct_to_obj_ids.items():
            ct = ContentType.objects.get_for_id(ct_id)
            model = ct.model_class()
            objs = model.objects.filter(id__in=obj_ids)
            ct_to_obj_id_to_obj[ct_id] = dict([(o.pk, o) for o in objs])
        # Put objects back in activity
        ret = []
        for activity in qs:
            try:
                for field in fields:
                    ct_id = getattr(activity, field + '_ct_id')
                    obj_id = getattr(activity, field + '_id')
                    if ct_id and obj_id:
                        obj = ct_to_obj_id_to_obj[ct_id].get(obj_id)
                        if not obj:
                            raise ObjectDoesNotExist
                        setattr(activity, field, obj)
                ret.append(activity)
            except ObjectDoesNotExist:
                pass
        return ret

    def filter_active(self, activities):
        """
        activities = list, not a queryset
        """
        ret = []
        for activity in activities:
            if activity.wrapper().is_active():
                ret.append(activity)
        return ret


@python_2_unicode_compatible
class Activity(models.Model):
    objects = ActivityManager()

    actor_ct = models.ForeignKey(ContentType, related_name='+',
                                 blank=True, null=True)
    actor_id = models.PositiveIntegerField()
    actor = GenericForeignKey('actor_ct', 'actor_id')

    target_ct = models.ForeignKey(ContentType, related_name='+',
                                  blank=True, null=True)
    target_id = models.PositiveIntegerField()
    target = GenericForeignKey('target_ct', 'target_id')

    action_object_ct = models.ForeignKey(ContentType, related_name='+',
                                         blank=True, null=True)
    action_object_id = models.PositiveIntegerField()
    action_object = GenericForeignKey(
        'action_object_ct',
        'action_object_id')
    type = models.CharField(max_length=100,
                            choices=registry.as_choices(),
                            verbose_name=_("type"))
    flags = models.BigIntegerField(default=0)  # Project specific flags
    extra_data = models.TextField(
        _("additional data"),
        blank=True)
    created_at = models.DateTimeField(_("created at"), auto_now_add=True)

    class Meta:
        verbose_name = _('activity')
        verbose_name_plural = _('activities')

    def get_type(self):
        activity_type = registry.by_id(self.type)
        return activity_type

    def wrapper(self):
        return self.get_type().Wrapper(self)

    def __str__(self):
        return self.type


class NoticeManager(models.Manager):

    def send(self, activity, users, request=None):
        for user in users:
            notice = Notice.objects.create(
                user=user,
                activity=activity,
                created_at=activity.created_at)
            activity.wrapper().send_notice_mail(notice, request)


class Notice(models.Model):
    objects = NoticeManager()
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        related_name='+',
        verbose_name=_("user"))
    activity = models.ForeignKey(Activity)
    # Denormalized.. matches activity
    created_at = models.DateTimeField(_('created at'))
    read_at = models.DateTimeField(_("read at"), blank=True, null=True)
    archived_at = models.DateTimeField(_("archived at"), blank=True, null=True)

    class Meta:
        verbose_name = _('notice')
        verbose_name_plural = _('notices')
        index_together = [
            ['user', 'read_at']
        ]

registry.load()



import json
from django.utils import six

from actistream import registry
from .adapter import get_adapter


class ActivityTypeMeta(type):
    def __new__(mcs, name, bases, d):
        userdef = bases[0] != object
        if userdef:
            d['app'] = d['__module__'].split('.')[-2]
            d['code'] = name.lower()
            d['id'] = d['app'] + '.' + d['code']
        ret = type.__new__(mcs, name, bases, d)
        if userdef:
            registry.register(ret)
        return ret


class ActivityWrapper(object):

    def __init__(self, activity):
        self.activity = activity

    def __getattr__(self, name):
        aliases = self.activity.get_type().aliases
        if name in aliases:
            return getattr(self.activity, aliases[name])
        return self.__dict__[name]

    def get_context_data(self):
        return {'activity': self.activity}

    def is_active(self):
        return True

    def get_user_email(self, user):
        return user.email

    def get_notice_mail_context_data(self, notice, request=None):
        return {
            'activity': self.activity,
            'activity_context':
            self.get_context_data(),
            'notice': notice,
            'user': notice.user}

    def render_notice_mail(self, notice, request=None):
        ctx = self.get_notice_mail_context_data(notice, request)
        activity_type = self.activity.get_type()
        return get_adapter(request).render_mail(
            '{0}/activities/{1}'
            .format(activity_type.app,
                    activity_type.code),
            self.get_user_email(notice.user),
            context=ctx)

    def send_notice_mail(self, notice, request=None):
        msg = self.render_notice_mail(notice, request)
        msg.send()
        return msg


@six.add_metaclass(ActivityTypeMeta)
class ActivityType(object):
    aliases = {}
    Wrapper = ActivityWrapper

    @classmethod
    def create(cls, **kwargs):
        from .models import Activity
        commit = kwargs.pop('commit', True)
        activity_kwargs = {}
        for k, v in kwargs.items():
            activity_kwargs[cls.aliases.get(k, k)] = v
        extra_data = activity_kwargs.get('extra_data')
        if extra_data is not None and not isinstance(
                extra_data, six.string_types):
            activity_kwargs['extra_data'] = json.dumps(extra_data)
        activity = Activity(type=cls.id,
                            **activity_kwargs)
        if commit:
            activity.save()
        return activity



from actistream.types import ActivityWrapper, ActivityType


class CommentPosted(ActivityType):
    verbose_name = 'Comment posted'

    class Wrapper(ActivityWrapper):

        def get_action_url(self):
            return self.activity.action_object.get_absolute_url()



# -*- coding: utf-8 -*-
# Generated by Django 1.9.8 on 2016-09-15 20:40
from __future__ import unicode_literals

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Activity',
            options={'verbose_name': 'activity', 'verbose_name_plural': 'activities'},
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('actor_id', models.PositiveIntegerField()),
                ('target_id', models.PositiveIntegerField()),
                ('action_object_id', models.PositiveIntegerField()),
                ('type', models.CharField(max_length=100, verbose_name='type')),
                ('flags', models.BigIntegerField(default=0)),
                ('extra_data', models.TextField(blank=True, verbose_name='additional data')),
                ('created_at', models.DateTimeField(auto_now_add=True, verbose_name='created at')),
                ('action_object_ct', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='+', to='contenttypes.ContentType')),
                ('actor_ct', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='+', to='contenttypes.ContentType')),
                ('target_ct', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='+', to='contenttypes.ContentType')),
            ],
        ),
        migrations.CreateModel(
            name='Notice',
            options={'verbose_name': 'notice', 'verbose_name_plural': 'notices'},
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(verbose_name='created at')),
                ('read_at', models.DateTimeField(blank=True, null=True, verbose_name='read at')),
                ('archived_at', models.DateTimeField(blank=True, null=True, verbose_name='archived at')),
                ('activity', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='actistream.Activity')),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to=settings.AUTH_USER_MODEL, verbose_name='user')),
            ],
        ),
    ]



from django.conf import settings
from django.template.loader import render_to_string
from django.contrib.sites.shortcuts import get_current_site
from django.core.mail import EmailMultiAlternatives, EmailMessage
from django.template import TemplateDoesNotExist

try:
    from django.utils.encoding import force_text
except ImportError:
    from django.utils.encoding import force_unicode as force_text

from .utils import import_attribute


class DefaultAdapter(object):

    def __init__(self, request=None):
        self.request = request

    def get_current_site(self):
        if self.request:
            site = get_current_site(self.request)
        elif ('django.contrib.sites' in settings.INSTALLED_APPS and
              settings.SITE_ID):
            from django.contrib.sites.models import Site
            site = Site.objects.get_current()
        else:
            site = None
        return site

    def format_email_subject(self, subject):
        prefix = getattr(settings, 'ACTISTREAM_EMAIL_SUBJECT_PREFIX', None)
        if prefix is None:
            site = self.get_current_site()
            if site:
                prefix = "[{name}] ".format(name=site.name)
            else:
                prefix = ''
        return prefix + force_text(subject)

    def format_email_body(self, body, ext):
        return body

    def get_from_email(self):
        """
        This is a hook that can be overridden to programatically
        set the 'from' email address for sending emails
        """
        return settings.DEFAULT_FROM_EMAIL

    def render_mail(self, template_prefix, email, context):
        """
        Renders an e-mail to `email`.  `template_prefix` identifies the
        e-mail that is to be sent, e.g. "account/email/email_confirmation"
        """
        context = dict(context)
        context['site'] = self.get_current_site()
        subject = render_to_string('{0}_subject.txt'.format(template_prefix),
                                   context)
        # remove superfluous line breaks
        subject = " ".join(subject.splitlines()).strip()
        subject = self.format_email_subject(subject)

        from_email = self.get_from_email()

        bodies = {}
        for ext in ['html', 'txt']:
            try:
                template_name = '{0}_message.{1}'.format(template_prefix, ext)
                bodies[ext] = self.format_email_body(
                    render_to_string(template_name, context).strip(),
                    ext)
            except TemplateDoesNotExist:
                if ext == 'txt' and not bodies:
                    # We need at least one body
                    raise
        if 'txt' in bodies:
            msg = EmailMultiAlternatives(subject,
                                         bodies['txt'],
                                         from_email,
                                         [email])
            if 'html' in bodies:
                msg.attach_alternative(bodies['html'], 'text/html')
        else:
            msg = EmailMessage(subject,
                               bodies['html'],
                               from_email,
                               [email])
            msg.content_subtype = 'html'  # Main content is now text/html
        return msg

    def send_mail(self, template_prefix, email, context):
        msg = self.render_mail(template_prefix, email, context)
        msg.send()


def get_adapter(request=None):
    path = getattr(
        settings, 'ACTISTREAM_ADAPTER', 'actistream.adapter.DefaultAdapter')
    return import_attribute(path)(request)



# -*- coding: utf-8 -*-
# Generated by Django 1.10.5 on 2017-02-27 15:50
from __future__ import unicode_literals

from django.conf import settings
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('actistream', '0001_initial'),
    ]

    operations = [
        migrations.AlterIndexTogether(
            name='notice',
            index_together=set([('user', 'read_at')]),
        ),
    ]



from django.template import Library
from django.template.loader import render_to_string

from ..models import Activity

register = Library()


@register.simple_tag(takes_context=True)
def render_activity(context, activity):
    activity_type = activity.get_type()
    data = activity.wrapper().get_context_data()
    return render_to_string('{0}/activities/{1}_detail.html'
                            .format(activity_type.app,
                                    activity_type.code),
                            data,
                            context)


@register.assignment_tag
def renderable_activities(activities):
    a = Activity.objects.fetch_related(activities)
    return Activity.objects.filter_active(a)



from importlib import import_module

VERSION = (1, 0, 5, 'final', 0)

__title__ = 'django-actistream'
__version_info__ = VERSION
__version__ = '.'.join(map(str, VERSION[:3])) + ('-{}{}'.format(
    VERSION[3], VERSION[4] or '') if VERSION[3] != 'final' else '')
__author__ = 'Raymond Penners'
__license__ = 'MIT'
__copyright__ = 'Copyright 2017 Raymond Penners and contributors'


class ActivityTypeRegistry(object):
    def __init__(self):
        self.type_map = {}

    def get_list(self):
        return self.type_map.values()

    def register(self, cls):
        self.type_map[cls.id] = cls()

    def by_id(self, id):
        return self.type_map[id]

    def as_choices(self):
        for nt in self.get_list():
            yield (nt.id, nt.verbose_name)

    def load(self):
        from django.conf import settings
        for app in settings.INSTALLED_APPS:
            try:
                m = app + '.activities'
                import_module(m)
            except ImportError:
                pass


registry = ActivityTypeRegistry()
